#--------------
#	R file to generate the 3D plots from the alpha beta grid of the CoHSI equation.
#
#	Author:	Les Hatton, with grateful thanks to Karline Soetaert and many other
#	contributors to the R project and its associated graphics support.
#
#	Revision:	$Revision: 1.1 $
#	Date:	$Date: 2018/07/01 16:53:40 $
#--------------
##	Pull in the necessary libraries.  If not present, use install.packages("tgp") etc.
#
library(tgp)
library(akima)
library(plot3D)
library(reshape2)
#
#	We have one csv file created by gridAlphaBeta.sh with columns
#	alpha,beta,mean,median,mode
#
#	First we create the full data frames.
#
fullabframe <- read.csv(file="gridAlphaBeta.csv", header=TRUE, sep=",")
#
#	The colnames are given by
#
colnames(fullabframe)
#	[1] "alpha"  "beta"   "mean"   "median" "mode"
#
#	Now we subframe out each of mean, median, mode.
#
abmean <- fullabframe[c("alpha","beta","mean")]
abmedian <- fullabframe[c("alpha","beta","median")]
abmode <- fullabframe[c("alpha","beta","mode")]
#
#	To illustrate
#
colnames(abmean)
#	[1] "alpha"  "beta"   "mean"
#
#	We now coerce these into matrices using the acast() function from
#	the reshape2 library.
#
mabmean <- acast(abmean,alpha~beta, value.var="mean")
mabmedian <- acast(abmedian,alpha~beta, value.var="median")
mabmode <- acast(abmode,alpha~beta, value.var="mode")
#
#	These are same for each plot so just use mabmean.
#
x <- seq(3, 6, length.out = nrow(mabmean))
y <- seq(0, 1, length.out = ncol(mabmean))
#--------------
##	Plot of mean with accompanying contour.
#
postscript(file = "meanAlphaBeta.eps", horizontal = FALSE)
par(mfrow = c(1,1))
persp3D( x, y, z = mabmean, zlim = c(-500, 750), phi = 20, bty = "b",   
	colkey = list(length = 0.2, width = 0.4, shift = 0.15, cex.axis = 0.8, cex.clab = 0.85), 
	lighting = FALSE, contour = TRUE, theta = 60, ticktype = "detailed",
	xlab = "alpha", ylab = "beta", zlab = "mean",
#	main = "Mean plotted as a function of (alpha,beta)",
	clab = c("mean"), plot = FALSE)

image3D( x, y, z = -500, colvar = mabmean, add = TRUE, resfac = c(4,4),
	colkey = FALSE, contour = TRUE,
	clab = c("alpha","beta"),
	plot = TRUE)
dev.off()
#--------------
##	Plot of median with accompanying contour.
#
postscript(file = "medianAlphaBeta.eps", horizontal = FALSE)
par(mfrow = c(1,1))
persp3D(x, y, z = mabmedian, zlim = c(-150, 150), phi = 20, bty = "b",   
	colkey = list(length = 0.2, width = 0.4, shift = 0.15, cex.axis = 0.8, cex.clab = 0.85), 
	lighting = FALSE, contour = TRUE, theta = 60, ticktype = "detailed",
	xlab = "alpha", ylab = "beta", zlab = "median",
#	main = "Median plotted as a function of (alpha,beta)",
	clab = c("median"), plot = FALSE)

image3D(x, y, z = -150, colvar = mabmedian, add = TRUE, resfac = c(4,4),
	colkey = FALSE, contour = TRUE,
	clab = c("alpha","beta"),
	plot = TRUE)
dev.off()
#--------------
##	Plot of mode with accompanying contour.
#
postscript(file = "modeAlphaBeta.eps", horizontal = FALSE)
par(mfrow = c(1,1))
persp3D(x, y, z = mabmode, zlim = c(-50, 50), phi = 20, bty = "b",   
	colkey = list(length = 0.2, width = 0.4, shift = 0.15, cex.axis = 0.8, cex.clab = 0.85), 
	lighting = FALSE, contour = TRUE, theta = 60, ticktype = "detailed",
	xlab = "alpha", ylab = "beta", zlab = "mode",
#	main = "Mode plotted as a function of (alpha,beta)",
	clab = c("mode"), plot = FALSE)

image3D(x, y, z = -50, colvar = mabmode, add = TRUE, resfac = c(4,4),
	colkey = FALSE, contour = TRUE,
	clab = c("alpha","beta"),
	plot = TRUE)
dev.off()
#--------------

#	par("ask" = FALSE)
