#!/usr/bin/perl
#
#	Calculate the N(ti,ai;ai) function by recursion.
#
#	Revision:	$Revision: 1.12 $
#	Date:	$Date: 2018/07/01 16:28:09 $
#--------------------------------------------------------------
#
use 	strict;
use 	warnings;
use	Getopt::Long;
use	Math::GammaFunction qw/:all/;

my	$ALPHA	= 6.0;
my	$BETA	= 3.0;
my	$GAMMA	= 2.0;
my	$EPSILON	= -10.0;
my	$TIMIN	= 1;
my	$TIMAX	= 100;
my	$TIMAXFULL= 100;	# Overflow beyond this.

my	$AIMIN	= 1;
my	$AIMAX	= 50;

my	%opt		= ();

my	@pdfaipwr	= ();
my	@pdfai	= ();
my	@pdfdef	= ();
my	@pdfti	= ();

my	%d0pdfti	= ();

GetOptions(
	'alpha=f'		=> \$ALPHA,
	'beta=f'		=> \$BETA,
	'gamma=f'		=> \$GAMMA,
	'epsilon=f'	=> \$EPSILON,
	'd'    		=> \$opt{d},			# Generate defect model
	'dd'    		=> \$opt{dd},			# Generate defect distribution d0,..,d9
	'ddd'    		=> \$opt{ddd},			# Generate defect density distribution
	'dddd'  		=> \$opt{dddd},		# Generate defect density v. ti.
	'ddddd'  		=> \$opt{ddddd},		# Generate defect distribution d0 v. ti
	'f'    		=> \$opt{f},			# Generate normal full distribution
	'p'    		=> \$opt{p},
	's'    		=> \$opt{s},
	'timin=i'     	=> \$TIMIN,
	'timax=i'     	=> \$TIMAX,
	'v'     		=> \$opt{v},
	'vv'     		=> \$opt{vv},
	'verbose'     	=> \$opt{v},
);
#---------------------------------------------------
#	General purpose integer factorial function.
#	Computed factor by factor to delay onset of overflow.
#
sub	factorial()
{
	my	($n)		= @_;

	my	$result	= 1;
	for( my $i = 2; $i <= $n; ++$i )
	{
		$result	*= $i;
	}

	return	$result;
}
#---------------------------------------------------
#	The defect ti relationship. This is the integrated form of
#	log ti = -gamma -epsilon d(di)/dti.
#
sub	getDefects
{
	my	($ti,$epsilon,$gamma)	= @_;

	my	$di	= (($gamma - 1.0) * (1 - $ti) - $ti * log($ti)) / $epsilon;
	
	return	$di;
}
#---------------------------------------------------
#	The real-valued nCr operator using the gamma() function.
#
sub	nCrReal
{
	my	($n,$r)	= @_;
#
#	Calculate this to minimise round-off
#
	my	$result	= gamma($n+1) / (gamma($r+1)*gamma($n-$r+1));

	print "n = $n, r = $r, nCr = $result\n" if $opt{vv};

	return	$result;	
}
#---------------------------------------------------
#	The lower bound for N(ti,ai;ai), (ti!)/((ti/ai)!)^(ai)
#	using integer decomposition.
#
sub	calcLowerN
{
	my	($ti,$ai)	= @_;

	my	$remainder	= $ti % $ai;
	my	$m			= ($ti - $remainder) / $ai;
	my	$factm		= &factorial($m);
	my	$result		= &factorial($ti) / (($factm**($ai-1)) * &factorial($m+$remainder));

	return	$result;
}
#---------------------------------------------------
#	Calculates the full N(ti,ai;ai) using recursion as described in the arXiv paper.
#	Here we define this to be valid for real $ti and integer $ai since the recursion
#	does not depend on looping on $ti.
#
sub	calcFullN
{
	my	($ti,$ai)	= @_;

	my	%elements	= ();
	my	$status	= 0;

	$elements{1 . "-" . 1}	= 1;
	my	$lastai	= 1;

	for( my $i = 2; $i <= $ai; ++$i )
	{
		for( my $j = 1; $j < $i; ++$j )
		{
			$elements{$i . "-" . $j}	= &nCrReal($i,$j) * $elements{$j . "-" . $j};

			my	$value		= $elements{$i . "-" . $j};
			print "---> N($ti,$i;$j) = &nCrReal($ai,$j) * N($ti,$j;$j) = $value\n" if $opt{vv};
		}

		$elements{$i . "-" . $i}	= ($i**$ti);

		print "N($ti,$i;$i) = ($i**$ti)\n" if $opt{vv};

		for( my $j = 1; $j < $i; ++$j )
		{
#
#			Decrement the contributions up to $i,$i.
#
			$elements{$i . "-" . $i}	-= $elements{$i . "-" . $j};
			my	$residual		= $elements{$i . "-" . $i};
			my	$value		= $elements{$i . "-" . $j};

			print " - N($ti,$i;$j) ($value) = $residual\n" if $opt{vv};
		}

		if ( $elements{$i . "-" . $i} < 0.0 )
		{
#
#			The decrementing has led to a negative number
#
			print " - IS BAD\n" if $opt{vv};
			$status	= 1;
			last;
		}

		$lastai	= $i;
	}

	if ( defined $elements{$lastai . '-' . $lastai} )
	{
		print "###> N($ti,$lastai;$lastai) (ai = $ai)  = $elements{$lastai . '-' . $lastai};\n" if $opt{v};
	}

	return	($status,$elements{$lastai . "-" . $lastai});
}
#---------------------------------------------------
#	Populate the N() data structure in the region in which it is defined.
#
my	%N			= ();
my	%dlogNdti		= ();

for( my $ti = $TIMIN; $ti <= $TIMAXFULL; ++$ti )
{
	my	$maxai	= $ti;
	if ( $ti > $AIMAX )		{ $maxai	= $AIMAX; }

	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	($status,$value)		= &calcFullN($ti, $ai);

		if ( $status == 0 )			{ $N{$ai . "-" . $ti}	= $value; }
	}
}
#---------------------------------------------------
#	N() now contains the chocolate box function where it can be calculated.
#	Calculate the derivatives.
#
my	$aisum	= 0.0;
my	$firstti	= 0;
my	$firstaii	= 0;
my	$lastti	= 0;

for( my $ti = $TIMIN; $ti <= $TIMAXFULL; ++$ti )
{
#
#	Incorporates Ramanujan's approximation for log n!.
#	https://en.wikipedia.org/wiki/Stirling%27s_approximation
#	Setting it equal to 0 gives the derivative of the simplest Stirling approximation d/dn (n*ln(n) - n) = log n.
#	Ramanujan's approximation is:-
#
#		log n! \approx n log n - n + 1/6 log(n + 4n^2 + 8n^3) + 1/2log(\pi)
#
#	Taking the derivative of this gives the srinivasa term below.
#
#	my	$srinivasa	= 0.0;
	my	$srinivasa	= (1.0 + 8*$ti + 24*$ti*$ti) / (6.0*($ti + 4*$ti*$ti + 8*$ti*$ti*$ti));
#
#	The CoHSI equation is:-
#
#		-(log($ti) + $srinivasa - $ALPHA)/$BETA = d/dti(log N($ti,$ai))
#
#	Here, N() is the chocolate box function and log N() is its HS information.
#
	my	$lhscohsi		= -(log($ti) + $srinivasa - $ALPHA)/$BETA;

	my	$tip3		= $ti + 3;
	my	$tip2		= $ti + 2;
	my	$tip1		= $ti + 1;
	my	$tim1		= $ti - 1;

	my	$maxai		= $ti;
	if ( $ti > $AIMAX )	{ $maxai	= $AIMAX; }
#
#	Define the straddle boundaries for $ai.
#
	my	$apihigh		= 1;
	my	$straddle		= 0;

	my	%allexprhscohsi	= ();
	my	%allexplhscohsi	= ();
#
#	Now compute the rhs of the CoHSI equation for a range of $ai for this $ti.
#
	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	$aip1	= $ai + 1;
#
#		This is the essence of the algorithm.  We must determine the gradient of the
#		chocolate box function for small $ti as accurately as possible.
#
		if ( $ai == 1 )
		{
#
#			log N() == 0 for all ti so d/dti = 0.
#
			$dlogNdti{$ai . "-" . $ti}	= 0.0;
		}
		elsif ( defined $N{$ai . "-" . $tim1} 
				&& defined $N{$ai . "-" . $ti} 
				&& defined $N{$ai . "-" . $tip1} 
				&& defined $N{$ai . "-" . $tip2} )
		{
#
#			Centred difference.
#
			if ( ($ti <= 10) && ($ai >= 2) )
			{
#
#				Use more accurate scheme for small $ti.
#
				my	$dti				= 0.01;
				my	($statuslo,$flow)	= &calcFullN($ti-$dti, $ai);
				my	($statushi,$fhigh)	= &calcFullN($ti+$dti, $ai);
				my	$dlogNdti			= 0.5*(log($fhigh) - log($flow)) / $dti;

				if ( ($statuslo == 0) && ($statushi == 0) )
				{
					$dlogNdti{$ai . "-" . $ti}	= $dlogNdti;
					print "fff> (ti=$ti,ai=$ai) EXPERIMENTAL d/dti = $dlogNdti\n" if $opt{v};
				}
				else
				{
					print "fff> statuslo = $statuslo, statushi = $statushi\n";
				}
			}
			else
			{
				$dlogNdti{$ai . "-" . $ti}	= 0.25 * (log($N{$ai . "-" . $tip1}) 
												- log($N{$ai . "-" . $tim1}) 
												+ log($N{$ai . "-" . $tip2}) 
												- log($N{$ai . "-" . $ti}));
			}

			print "fff> (ti=$ti,ai=$ai) CENTRAL d/dti = $dlogNdti{$ai . '-' . $ti}\n" if $opt{v};
		}
		elsif ( defined $N{$ai . "-" . $ti} 
				&& defined $N{$ai . "-" . $tip1} 
				&& defined $N{$ai . "-" . $tip2} 
				&& defined $N{$ai . "-" . $tip3} )
		{
#
#			Forward difference.
#
			if ( ($ti <= 10) && ($ai >= 2) )
			{
				my	$dti				= 0.01;
				my	($statuslo,$flow)	= &calcFullN($ti, $ai);
				my	($statushi,$fhigh)	= &calcFullN($ti+$dti, $ai);
				my	$dlogNdti			= (log($fhigh) - log($flow)) / $dti;

				if ( ($statuslo == 0) && ($statushi == 0) )
				{
					$dlogNdti{$ai . "-" . $ti}	= $dlogNdti;
					print "fff> (ti=$ti,ai=$ai) EXPERIMENTAL d/dti = $dlogNdti\n" if $opt{v};
				}
				else
				{
					print "fff> statuslo = $statuslo, statushi = $statushi\n";
				}
			}
			else
			{
				$dlogNdti{$ai . "-" . $ti}	= (log($N{$ai . "-" . $tip1}) 
											+ log($N{$ai . "-" . $tip2}) 
											+ log($N{$ai . "-" . $tip3}) 
											- 3.0*log($N{$ai . "-" . $ti}))/3.0;
			}

			print "fff> (ti=$ti,ai=$ai) FORWARD d/dti = $dlogNdti{$ai . '-' . $ti}\n" if $opt{v};
		}
		else
		{
#
#			Give up for now.
#
			$dlogNdti{$ai . "-" . $ti}    = 0.0;
			print "fff> (ti=$ti,ai=$ai) No d/dti\n" if $opt{v};
		}
#
#		Compute the rhs of the CoHSI for large $ti when N() = $ai^$ti.
#
		my	$rhspower		= log($ai);
		my	$exprhspower	= exp($rhspower);
#
#		Compute the rhs of the CoHSI for smaller $ti using the full solution.
#
		my	$rhscohsi		= $dlogNdti{$ai . "-" . $ti};
		my	$exprhscohsi	= exp($rhscohsi);

		my	$diff		= $rhscohsi - $rhspower;

		print "ppp> (ti=$ti,ai=$ai) Full - power = $diff\n" if $opt{v};
#
#		Get the lhs of the CoHSI equation for all $ti.
#
		my	$explhscohsi	= exp($lhscohsi);
#
#		Store rhs and lhs of the full equation for a range of $ai.
#
		$allexprhscohsi{$ai}	= $exprhscohsi;
		$allexplhscohsi{$ai}	= $explhscohsi;
#
#		Finally, calculate the following diagnostics.
#
		if ( defined $N{$ai . "-" . $ti} )
		{
			my	$lowerN		= &calcLowerN($ti,$ai);
			my	$fullN		= $N{$ai . "-" . $ti};
			my	$powerN		= $ai**$ti;

			my	$loglowerN	= log($lowerN);
			my	$logfullN		= log($fullN);
			my	$logpowerN	= log($powerN);
#
#			Various diagnostics.
#
			print "===> ti = $ti, ai = $ai, lhscohsi = $lhscohsi, rhs = $rhscohsi, explhscohsi = $explhscohsi, exprhscohsi = $exprhscohsi\n" if $opt{v};
#			print "     loglowerN = $loglowerN, logfullN = $logfullN, logpowerN = $logpowerN\n" if $opt{v};
		}
	}
#
#	Now interpolate.
#
#	To find the equivalent solution for the full equation, we must find which value of ai
#	forces log N(ti,ai;ai) = explhscohsi for this value ot ti.
#
	if ( $ti == 1 )
	{
#
#		We know (1,1) lies on the CoHSI distribution so add it as the first point.
#
		push @pdfaipwr, exp($ALPHA/$BETA);			# lhs when log(ti) = 0.
		push @pdfai, 1.0;
		push @pdfti, 1.0;
	}

	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	$aip1	= $ai + 1;
#
#		Step through the array of values of the rhs for each ai.
#
		if ( defined $allexprhscohsi{$ai} && defined $allexprhscohsi{$aip1} )
		{
			if ( ($allexprhscohsi{$ai} < $allexplhscohsi{$ai}) && ($allexprhscohsi{$aip1} >= $allexplhscohsi{$ai}) )
			{
#
#				Straddled with (ai,aip1), interpolate.
#
				my	$ai_interp	= $ai + ($allexplhscohsi{$ai} - $allexprhscohsi{$ai})/($allexprhscohsi{$aip1} - $allexprhscohsi{$ai});
				my	$tiaidiff		= $ti - $ai_interp;

				print "yyy> ti = $ti, ai = $ai, lhscohsi = $allexplhscohsi{$ai}, rhscohsi{$ai} = $allexprhscohsi{$ai}, allexprhscohsi{$aip1} = $allexprhscohsi{$aip1}, ai_interp = $ai_interp, tiaidiff = $tiaidiff\n" if $opt{v};
#				print "$ti $allexplhscohsi{$ai} $ai_interp\n";

				if ( $firstti == 0 )
				{
#
#					Set the first ti for which we have a solution.
#					We currently do not use this.  It could be used to interpolate the low
#					end but this is artificial.
#
					$firstti	= $ti;
					$firstaii	= $ai_interp;
				}

				$lastti	= $ti;

				push @pdfaipwr, $allexplhscohsi{$ai};
				push @pdfai, $ai_interp;
				push @pdfti, $ti;

				$aisum	+= $ai_interp;
			}
			else
			{
				print "xxx> ti = $ti, ai = $ai, allexplhscohsi{$ai} = $allexplhscohsi{$ai}, allexprhscohsi{$ai} = $allexprhscohsi{$ai}, allexprhscohsi{$aip1} = $allexprhscohsi{$aip1}\n" if $opt{v};
			}
		}
	}
}
#
#	Fill out the rest with the power-law as it is by now indistinguishable, so
#	dlogN(ti,ai)/dti = d/dti log(ai^ti) = log ai, so the CoHSI equation is now
#	-(log($ti) - $ALPHA)/$BETA = log ai;
#
for( my $ti = $lastti+1; $ti <= $TIMAX; ++$ti )
{
	my	$lhscohsi		= -(log($ti) - $ALPHA)/$BETA;
	my	$aipower		= exp($lhscohsi);

#	print "$ti $aipower $aipower\n";

	push @pdfaipwr, $aipower;
	push @pdfai, $aipower;
	push @pdfti, $ti;

	$aisum	+= $aipower;
}
#
#	Initialise the defect data structures.
#
my	$disum	= 0.0;
my	%diti	= ();
my	%didens	= ();
my	$d0		= 0.0;
my	$d1		= 0.0;
my	$d2		= 0.0;
my	$d3		= 0.0;
my	$d4		= 0.0;
my	$d5		= 0.0;
my	$d6		= 0.0;
my	$d7		= 0.0;
my	$d8		= 0.0;
my	$d9		= 0.0;

my	$d0sum	= 0.0;
#
#	Print out the ai v. ti CoHSI distribution in some form, normalised or not.
#
while(1)
{
	my	$ti		= shift(@pdfti);
	my	$ai		= shift(@pdfai);
	my	$aipwr	= shift(@pdfaipwr);
	my	$aibc	= $ti;
	my	$unnormai	= 0;

	if ( ! defined $ti )		{ last; }

	$unnormai		= $ai;

	if ( $opt{p} )
	{
		$ai		/= $aisum;
		$aipwr	/= $aisum;
		$aibc	/= $aisum;
	}

	print "$ti $aipwr $ai $aibc $unnormai\n" if $opt{f};
#
#	Create the di v. ti distribution.
#
#	$di is the probability of a token being in error in a component
#	of size $ti.  $ditot is the solution of the defect equation.
#
	my	$ditot	= &getDefects($ti,$EPSILON,$GAMMA);
	my	$di		= $ditot/$ti;
#
#	$d0-$d9 is the percentage of all components with 0,..,9
#	defects respectively.  We use the Poisson distribution.
#
	my	$thisd0	= $ai * ($di**0.0) * exp(-$di) / gamma(1);

	$d0pdfti{$ti}	= $thisd0;			# Zero-defect ti distribution.
#
#	Increment the accumulators.
#
	$d0		+= $thisd0;
	$d1		+= $ai * ($di**1.0) * exp(-$di) / gamma(2);
	$d2		+= $ai * ($di**2.0) * exp(-$di) / gamma(3);
	$d3		+= $ai * ($di**3.0) * exp(-$di) / gamma(4);
	$d4		+= $ai * ($di**4.0) * exp(-$di) / gamma(5);
	$d5		+= $ai * ($di**5.0) * exp(-$di) / gamma(6);
	$d6		+= $ai * ($di**6.0) * exp(-$di) / gamma(7);
	$d7		+= $ai * ($di**7.0) * exp(-$di) / gamma(8);
	$d8		+= $ai * ($di**8.0) * exp(-$di) / gamma(9);
	$d9		+= $ai * ($di**9.0) * exp(-$di) / gamma(10);
#
#	TO find solutions which satisfy CoHSI and the defect equation, multiply
#	the two solutions.
#
	my	$adi		= $ai * $ditot;
	print ">>>> $ti $ditot $ai $adi\n" if $opt{v};
#
#	Prepare the normalisation.
#
	$disum		+= $adi;
	$diti{$ti}	= $adi;
	$didens{$ti}	= $di;
}
#
#	Make sure they add up to 100%.
#
my	$d09sum	= 0.0;

$d09sum	+= $d0 + $d1 + $d2 + $d3 + $d4 + $d5 + $d6 + $d7 + $d8 + $d9;

$d0	= (100.0 * $d0) / $d09sum;
$d1	= (100.0 * $d1) / $d09sum;
$d2	= (100.0 * $d2) / $d09sum;
$d3	= (100.0 * $d3) / $d09sum;
$d4	= (100.0 * $d4) / $d09sum;
$d5	= (100.0 * $d5) / $d09sum;
$d6	= (100.0 * $d6) / $d09sum;
$d7	= (100.0 * $d7) / $d09sum;
$d8	= (100.0 * $d8) / $d09sum;
$d9	= (100.0 * $d9) / $d09sum;

print "0 $d0\n" if $opt{dd};
print "1 $d1\n" if $opt{dd};
print "2 $d2\n" if $opt{dd};
print "3 $d3\n" if $opt{dd};
print "4 $d4\n" if $opt{dd};
print "5 $d5\n" if $opt{dd};
print "6 $d6\n" if $opt{dd};
print "7 $d7\n" if $opt{dd};
print "8 $d8\n" if $opt{dd};
print "9 $d9\n" if $opt{dd};

if ( $opt{s} )
{
	print "\n\n";

	for ( my $ti = 0; $ti < 100; ++$ti )
	{
		my	$aibc	= $ti / $aisum;
		print "$ti $aibc\n";
	}
}

foreach	my	$tix	(sort { $a <=> $b } keys %diti)
{
#
#	Extract the normalised defect distribution.
#
	my	$di		= $diti{$tix} / $disum;
#
#	Extract the defect density distribution.
#
	my	$did		= $di / $tix;
	my	$didd	= $didens{$tix};
	my	$d0pdf	= $d0pdfti{$tix};
#	print "$tix $di $did\n" if $opt{d};
	print "$tix $di\n" if $opt{d};
	print "$tix $did\n" if $opt{ddd};
	print "$tix $didd\n" if $opt{dddd};
	print "$tix $d0pdf\n" if $opt{ddddd};
}

exit(0);
